"""
This file defines the interface between python and the Rust SurrealDB library for logging in.
"""
from typing import Dict, Optional

from surrealdb.rust_surrealdb import blocking_authenticate
from surrealdb.rust_surrealdb import blocking_sign_in
from surrealdb.rust_surrealdb import blocking_sign_up

from surrealdb.errors import SurrealDbError


class SignInMixin:
    """
    This class is responsible for the interface between python and the Rust SurrealDB library for logging in.
    """
    def signin(self: "SurrealDB", data: Optional[Dict[str, str]] = None) -> None:
        """
        Signs in to the database.

        :param password: the password to sign in with
        :param username: the username to sign in with

        :return: None
        """
        if data is None:
            data = dict()
        data = {key.lower(): value for key, value in data.items()}

        password: str = data.get("password", data.get("pass", data.get("p", "root")))
        username: str = data.get("username", data.get("user", data.get("u", "root")))

        try:
            blocking_sign_in(self._connection, password, username)
        except Exception as e:
            SurrealDbError(e)

    def signup(self: "SurrealDB", namespace: str, database: str, data: Optional[Dict[str, str]] = None) -> str:
        """
        Signs up to an auth scope within a namespace and database.

        :param namespace: the namespace the auth scope is associated with
        :param database: the database the auth scope is associated with
        :param data: the data to sign up with
        :return: an JWT for that auth scope
        """
        try:
            return blocking_sign_up(self._connection, data, namespace, database)
        except Exception as e:
            SurrealDbError(e)

    def authenticate(self: "SurrealDB", jwt: str) -> bool:
        """
        Authenticates a JWT.

        :param jwt: the JWT to authenticate
        :return: None
        """
        try:
            return blocking_authenticate(self._connection, jwt)
        except Exception as e:
            SurrealDbError(e)



"""
This file defines the interface between python and the Rust SurrealDB library for querying the database.
queries can be found in the link below:
https://github.com/surrealdb/surrealdb/blob/main/lib/tests/fetch.rs
"""
import json
from typing import List, Union

from surrealdb.rust_surrealdb import blocking_query
from surrealdb.rust_surrealdb import blocking_select
from surrealdb.errors import SurrealDbError


class QueryMixin:
    """
    This class is responsible for the interface between python and the Rust SurrealDB library for creating a document.
    """
    def query(self: "SurrealDB", query: str) -> List[dict]:
        """
        queries the database.

        :param query: the query to run on the database

        :return: None
        """
        try:
            return json.loads(blocking_query(self._connection, query))[0]
        except Exception as e:
            SurrealDbError(e)

    def select(self: "SurrealDB", resource: str) -> Union[List[dict], dict]:
        """
        Performs a select query on the database for a particular resource.

        :param resource: the resource to select from

        :return: the result of the select
        """
        return blocking_select(self._connection, resource)



"""
This file defines the interface between python and the Rust SurrealDB library for setting a key value.
"""
import json

from surrealdb.rust_surrealdb import blocking_set
from surrealdb.errors import SurrealDbError


class SetMixin:
    """
    This class is responsible for the interface between python and the Rust SurrealDB library for creating a document.
    """
    def set(self: "SurrealDB", key: str, value: dict) -> None:
        """
        Creates a new document in the database.

        :param name: the name of the document to create
        :param data: the data to store in the document

        :return: None
        """
        json_str = None
        try:
            json_str = json.dumps(value)
        except json.JSONEncodeError as e:
            print(f"cannot serialize value {type(value)} to json")
            SurrealDbError(e)
        if json_str is not None:
            try:
                blocking_set(self._connection, key, json_str)
            except Exception as e:
                SurrealDbError(e)



"""
This file defines the interface for the SurrealDB connection and operations on that connection.

# Usage
The connection can be used by the following code:
```python
from surrealdb.connection_interface import SurrealDB

connection = SurrealDB(url="ws://localhost:8080")
```
Existing connections can be used by the following code:
```python
from surrealdb.connection_interface import SurrealDB

connection = SurrealDB(url="ws://localhost:8080", existing_connection_id="some_connection_id")
```
"""
import uuid
from typing import Optional

from surrealdb.rust_surrealdb import blocking_make_connection
from surrealdb.rust_surrealdb import blocking_use_namespace
from surrealdb.rust_surrealdb import blocking_use_database

from surrealdb.execution_mixins.auth import SignInMixin

# import the mixins for operations for the connection
from surrealdb.execution_mixins.create import CreateMixin
from surrealdb.execution_mixins.query import QueryMixin
from surrealdb.execution_mixins.set import SetMixin
from surrealdb.execution_mixins.update import UpdateMixin


class ConnectionController(type):

    instances = {}
    main_connection = None

    def remove_connection(cls, connection_id: str) -> None:
        """
        Closes the connection with the given connection id.

        :param connection_id: the id of the connection to close

        :return: None
        """
        del cls.instances[connection_id]

    def __call__(cls, *args, **kwargs):

        # establish the main connection
        if kwargs.get("main_connection", False) is True:
            if cls.main_connection is None:
                instance = super(ConnectionController, cls).__call__(*args, **kwargs)
                cls.main_connection = instance
            return cls.main_connection
        if len(args) > 4 and args[3] is True:
            if cls.main_connection is None:
                instance = super(ConnectionController, cls).__call__(*args, **kwargs)
                cls.main_connection = instance
            return cls.main_connection

        # get the existing connection
        if kwargs.get("existing_connection_id", None) is not None:
            return cls.instances[kwargs["existing_connection_id"]]
        if len(args) > 2 and args[2] is not None and args[2] in cls.instances:
            return cls.instances[args[2]]

        # store the connection
        if kwargs.get("keep_connection", False) is True:
            instance = super(ConnectionController, cls).__call__(*args, **kwargs)
            cls.instances[instance.id] = instance
            return instance
        if len(args) > 1 and args[1] is True:
            instance = super(ConnectionController, cls).__call__(*args, **kwargs)
            cls.instances[instance.id] = instance
            return instance

        return super(ConnectionController, cls).__call__(*args, **kwargs)


class SurrealDB(
    CreateMixin,
    SignInMixin,
    SetMixin,
    QueryMixin,
    UpdateMixin,
    metaclass=ConnectionController
):
    """
    This class is responsible for managing the connection to SurrealDB and managing operations on the connection.
    """
    def __init__(self,
                 url: Optional[str] = None,
                 keep_connection: Optional[bool] = False,
                 existing_connection_id: Optional[str] = None,
                 main_connection: Optional[bool] = False
                 ) -> None:
        """
        The constructor for the SurrealDB class.

        :param url: the url to connect to SurrealDB with
        :param keep_connection: wether or not to keep the connection open after this object is destroyed
        :param existing_connection_id: the existing connection id to use instead of making a new connection
        """
        self._connection: Optional[str] = self._make_connection(url=url)
        self.id: str = str(uuid.uuid4()) if existing_connection_id is None else existing_connection_id
        self.keep_connection: bool = keep_connection
        self.main_connection: bool = main_connection

    def _make_connection(self, url: str) -> str:
        """
        Makes a connection to SurrealDB or establishes an existing connection.

        :param url: the url to connect to SurrealDB with
        :param existing_connection_id: the existing connection id to use instead of making a new connection
        :return: the connection id of the connection
        """
        return blocking_make_connection(url)

    def use_namespace(self, namespace: str) -> None:
        """
        Uses the given namespace in the connection.

        :param namespace: the namespace to use
        :return: None
        """
        blocking_use_namespace(self._connection, namespace)

    def use_database(self, database: str) -> None:
        """
        Uses the given database in the connection.

        :param database: the database to use
        :return: None
        """
        blocking_use_database(self._connection, database)



#!/usr/bin/env python
import pathlib

from setuptools import setup
from setuptools_rust import Binding, RustExtension


with open("README.md", "r") as fh:
    long_description = fh.read()


with open(str(pathlib.Path(__file__).parent.absolute()) + "/surrealdb/VERSION.txt", "r") as fh:
    version = fh.read().split("=")[1].replace("'", "")


setup(
    name="surrealdb-beta",
    author="Maxwell Flitton",
    author_email="maxwell@gmail.com",
    description="Calculates a Fibonacci number",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version,
    rust_extensions=[RustExtension("surrealdb.rust_surrealdb", binding=Binding.PyO3)],
    packages=[
        "surrealdb",
        "surrealdb.execution_mixins"
    ],
    package_data={
        "surrealdb": ["binaries/*"],
    },
    # rust extensions are not zip safe, just like C-extensions.
    zip_safe=False,
    include_package_data=True
)



"""
This file defines the interface between python and the Rust SurrealDB library for creating a document.
"""
import json
from typing import List, Union

from surrealdb.rust_surrealdb import blocking_create
from surrealdb.rust_surrealdb import blocking_delete

from surrealdb.errors import SurrealDbError


class CreateMixin:
    """
    This class is responsible for the interface between python and the Rust SurrealDB library for creating a document.
    """
    def create(self: "SurrealDB", name: str, data: dict) -> None:
        """
        Creates a new document in the database.

        :param name: the name of the document to create
        :param data: the data to store in the document

        :return: None
        """
        try:
            blocking_create(self._connection, name, json.dumps(data))
        except Exception as e:
            SurrealDbError(e)

    def delete(self: "SurrealDB", name: str) -> Union[List[dict], dict]:
        """
        Deletes a document in the database.

        :param name: the name of the document to delete

        :return: the record or records that were deleted
        """
        try:
            return blocking_delete(self._connection, name)
        except Exception as e:
            SurrealDbError(e)



"""
This file defines the interface between python and the Rust SurrealDB library for updating rows in the database. 
"""
import json
from typing import List, Union

from surrealdb.rust_surrealdb import blocking_merge
from surrealdb.rust_surrealdb import blocking_update
from surrealdb.rust_surrealdb import blocking_patch

from surrealdb.errors import SurrealDbError


class UpdateMixin:
    """
    This class is responsible for the interface between python and the Rust SurrealDB library for creating a document.
    """
    def update(self: "SurrealDB", resource: str, data: dict) -> Union[List[dict], dict]:
        """
        Updates the given resource with the given data.

        :param resource: the resource to update
        :param data: the data to update the resource with
        :return: the updated resource such as an individual row or a list of rows
        """
        try:
            return json.loads(blocking_update(self._connection, resource, data))
        except Exception as e:
            SurrealDbError(e)

    def merge(self: "SurrealDB", resource: str, data: dict) -> Union[List[dict], dict]:
        """
        Merges the given resource with the given data.

        :param resource: the resource to update
        :param data: the data to merge the resource with
        :return: the updated resource such as an individual row or a list of rows
        """
        try:
            return json.loads(blocking_merge(self._connection, resource, data))
        except Exception as e:
            SurrealDbError(e)

    def patch(self: "SurrealDB", resource: str, data: dict) -> Union[List[dict], dict]:
        """
        Patches the given resource with the given data.

        :param resource: the resource to update
        :param data: the data to patch the resource with
        :return: the updated resource such as an individual row or a list of rows
        """
        try:
            return json.loads(blocking_patch(self._connection, resource, data))
        except Exception as e:
            SurrealDbError(e)
